use crate::db::Dao;
use crate::errors::Error;
use crate::models::characters::Character;
use crate::models::proto::cofd::api::cofd_api_server::CofdApi;
use crate::models::proto::cofd::api::UpdateSkillValueRequest;
use crate::models::proto::cofd::cofd_sheet::Skill;
use crate::models::proto::cofd::*;
use crate::models::users::User;
use std::collections::btree_map::{Entry, OccupiedEntry};
use tonic::{Request, Response, Status};

/// Load the character belonging to the given user, as long as they're
/// the owner of that character. Returns an error if user is not
/// logged in, the owner of the character is not found, or the logged
/// in user does not have the permission to access this character.
async fn load_character(
    conn: &sqlx::SqlitePool,
    logged_in_user: Option<&User>,
    owner: &str,
    character_id: i32,
) -> Result<Character, Error> {
    let logged_in_user = logged_in_user.ok_or(Error::NotLoggedIn)?;

    let character: Character = conn
        .load_character(character_id)
        .await?
        .ok_or(Error::NotFound)?;

    if &logged_in_user.username != owner {
        return Err(Error::NoPermission);
    }

    Ok(character)
}

fn find_skill_entry<'a>(
    sheet: &'a mut CofdSheet,
    skill_name: &'a str,
) -> Option<OccupiedEntry<'a, String, Skill>> {
    let all_skills = vec![
        &mut sheet.mental_skills,
        &mut sheet.physical_skills,
        &mut sheet.social_skills,
    ];

    // Search all skill lists for this value using "workaround" to
    // break value from for loops.
    let skill: Option<OccupiedEntry<_, _>> = 'l: loop {
        for skill_map in all_skills {
            if let Entry::Occupied(entry) = skill_map.entry(skill_name.to_owned()) {
                break 'l Some(entry);
            }
        }

        break None;
    };

    skill
}

fn find_skill<'a>(sheet: &'a mut CofdSheet, skill_name: &'a str) -> Option<&'a mut Skill> {
    find_skill_entry(sheet, skill_name).map(|entry| entry.into_mut())
}

#[derive(Debug)]
pub struct CofdApiService {
    pub db: sqlx::SqlitePool,
}

#[tonic::async_trait]
impl CofdApi for CofdApiService {
    async fn update_skill_value(
        &self,
        request: Request<UpdateSkillValueRequest>, // Accept request of type HelloRequest
    ) -> Result<Response<Skill>, Status> {
        //Can use metadata map to add user id inside interceptor for auth.
        let request = request.into_inner();
        let mut character = load_character(
            &self.db,
            logged_in_user,
            &request.character_username,
            request.character_id,
        )
        .await?;

        let mut sheet: CofdSheet = character.try_deserialize()?;
        let skill: Option<&mut Skill> = find_skill(&mut sheet, &request.skill_name);

        skill
            .map(|s| s.dots = request.skill_value)
            .ok_or(Error::InvalidInput)?;

        println!("updated skill value",);

        character.update_data(sheet)?;
        self.db
            .update_character_sheet(&character)
            .await
            .map_err(|e| e.into())?; //TODO maybe use crate Error for db

        let reply = Skill::default();

        Ok(Response::new(reply)) // Send back our formatted greeting
    }
}
